#' @title ListPairwiseDistance
#' @name ListPairwiseDistance
#' @description Helper function to get all-vs-all repeat distances based on
#' RepeatAlignment.
#' @param x \code{list} of repeats as \code{DNAStringSet}
#' @param dist.type distance calculation method if no gendist matrix is
#' supplied [default: hamming]
#' @param dna.model a character string specifying the evolutionary model to be
#' used (see \code{ape::dist.dna}) [default: K80]
#' @param wmut weight for nucleotide point mutations [default: 1]
#' @param windel weight for unit insertions/deletions [default: 3.5]
#' @param wslippage weight for single unit duplications [default: 1.75]
#' @param exclude.pos positions of the repeat that will be excluded before
#' distance calculation [default: NULL]
#' @param post.include boolean if excluded positions should be added back for
#' the sequence alignment after distance calculation [default: FALSE]
#' @param output.dist specify if either the \code{distance} or
#' \code{sequence.dist} should be returned (see
#' \code{repeatR::RepeatAlignment}) [default: distance]
#' @param threads number of parallel threads [default: 1]
#' @import foreach
#' @import doMC
#' @importFrom Biostrings DNAString DNAStringSet AAString AAStringSet
#' readDNAStringSet readAAStringSet writeXStringSet width subseq translate
#' @importFrom ape dist.dna bionj write.tree as.DNAbin
#' @seealso \code{\link[ape]{dist.dna}}
#' @references Vara C et al. (2019) PRDM9 Diveristy at Fine Geographical Scale
#' Reveals Contrasting Evolutionary Patterns and Functional Constraints in
#' Natural Populations of House Mice. \emph{Molecular Biology and Evolution},
#' \bold{36(8)}, 1686-1700.
#' @examples
#' ##load example sequence data
#' data("mousePRDM9", package="repeatR")
#' myRepPattern<-"PY"
#' myRepLength<-84
#' mousePRDM9.random<-sample(mousePRDM9, 20)
#' mousePRDM9.random.split<-repeatR::splitRepByPattern(mousePRDM9.random,
#'     myRepPattern, myRepLength)
#' ##get distance for all-vs-all comparison excluding highly variable sites
#' dist.mat.hamming.exclude.pos<-repeatR::ListPairwiseDistance(
#'     x=mousePRDM9.random.split$cds,
#'     dist.type="hamming",
#'     wmut=1,
#'     windel=3.5,
#'     wslippage=1.75,
#'     exclude.pos=c(37:39,46:48,55:57),
#'     post.include=FALSE,
#'     output.dist="distance")
#' ##calculate bionj tree from resulting distances and write tree in newick
#' ##format
#' mousePRDM9.random.bionj<-ape::bionj(as.dist(dist.mat.hamming.exclude.pos))
#' ape::write.tree(mousePRDM9.random.bionj)
#' plot(mousePRDM9.random.bionj)
#' @export ListPairwiseDistance
#' @author Kristian K Ullrich

ListPairwiseDistance<-function(x, dist.type="hamming", dna.model="K80", wmut=1,
    windel=3.5, wslippage=1.75, exclude.pos=NULL, post.include=FALSE,
    output.dist="distance", threads=1){
    #check if all repeat units have the same length
    stopifnot(length(unique(unlist(lapply(x, function(x) unique(
        width(x))))))==1)
    x.repeat.units <- unlist(lapply(x, length))
    x.red <- x[x.repeat.units!=1]
    x.red.repeat.units <- unlist(lapply(x.red, length))
    x.ordered <- x.red[order(x.red.repeat.units, decreasing=TRUE)]
    doMC::registerDoMC(threads)
    i <- NULL
    j <- NULL
    out <- foreach::foreach(i=seq(from=1, to=length(x.ordered) - 1),
        .combine=c) %dopar% {
        foreach::foreach(j=seq(from=i + 1, to=length(x.ordered)),
            .combine=c) %do% {
                tmp.x <- x.ordered[[i]]
                tmp.x.name <- names(x.ordered)[i]
                tmp.y <- x.ordered[[j]]
                tmp.y.name <- names(x.ordered)[j]
                tmp.alg <- repeatR::RepeatAlignment(tmp.x, tmp.y,
                    dist.type=dist.type, dna.model=dna.model, wmut=wmut,
                    windel=windel, wslippage=wslippage,
                    exclude.pos=exclude.pos, post.include=post.include)
                if(output.dist=="distance"){
                    return(tmp.alg$distance)
                }
                if(output.dist=="sequence.dist"){
                    return(tmp.alg$sequence.dist)
                }
            }
    }
    dist.mat <- matrix(0, nrow=length(x.ordered), ncol=length(x.ordered))
    colnames(dist.mat) <- rownames(dist.mat) <- names(x.ordered)
    dist.mat[lower.tri(dist.mat,diag=FALSE)] <- out
    dist.mat <- as.matrix(as.dist(dist.mat))
    return(dist.mat)
}
