#' @title RepeatAlignment
#' @name RepeatAlignment
#' @description This function calculates genetic distances between complex
#' repeats as described in \emph{Vara (2019)}.
#' This function is forked from
#' \url{github.com/lucaferretti/RepeatDistance/blob/master/RepeatDistance.R}
#' and adapted to work with \code{Biostrings} objects.
#' @param seq1 \code{DNAStringSet}
#' @param seq2 \code{DNAStringSet}
#' @param seq1.name seq1 name [default: seq1]
#' @param seq2.name seq2 name [default: seq2]
#' @param gendist pairwise distance matrix of seq1 and seq2 repeats
#' [default: NULL]
#' @param dist.type distance calculation method if no gendist matrix is
#' supplied [default: hamming]
#' @param dna.model a character string specifying the evolutionary model to be
#' used (see \code{ape::dist.dna}) [default: K80]
#' @param wmut weight for nucleotide point mutations [default: 1]
#' @param windel weight for unit insertions/deletions [default: 3.5]
#' @param wslippage weight for single unit duplications [default: 1.75]
#' @param exclude.pos positions of the repeat that will be excluded before
#' distance calculation [default: NULL]
#' @param post.include boolean if excluded positions should be added back for
#' the sequence alignment after distance calculation [default: FALSE]
#' @importFrom Biostrings DNAString DNAStringSet AAString AAStringSet
#' readDNAStringSet readAAStringSet writeXStringSet width subseq
#' @importFrom ape dist.dna
#' @seealso \code{\link[ape]{dist.dna}}
#' @references Vara C et al. (2019) PRDM9 Diveristy at Fine Geographical Scale
#' Reveals Contrasting Evolutionary Patterns and Functional Constraints in
#' Natural Populations of House Mice. \emph{Molecular Biology and Evolution},
#' \bold{36(8)}, 1686-1700.
#' @examples
#' ##load example sequence data
#' data("mousePRDM9", package="repeatR")
#' myRepPattern<-"PY"
#' myRepLength<-84
#' mousePRDM9.random<-sample(mousePRDM9, 20)
#' mousePRDM9.random.split<-repeatR::splitRepByPattern(mousePRDM9.random,
#'     myRepPattern, myRepLength)
#' ##get repeat alignment
#' repeatR::RepeatAlignment(mousePRDM9.random.split$cds[[1]],
#'     mousePRDM9.random.split$cds[[2]])
#' ##get repeat alignment and exclude highly variable sites for the distance
#' ##calculation
#' repeatR::RepeatAlignment(mousePRDM9.random.split$cds[[1]],
#'     mousePRDM9.random.split$cds[[2]],
#'     exclude.pos = c(37:39, 46:48, 55:57))
#' ##get repeat alignment, exclude highly variable sites for the distance
#' ##calculation and add them back afterwards
#' repeatR::RepeatAlignment(mousePRDM9.random.split$cds[[1]],
#'     mousePRDM9.random.split$cds[[2]],
#'     exclude.pos = c(37:39, 46:48, 55:57),
#'     post.include = TRUE)
#' @export RepeatAlignment
#' @author Kristian K Ullrich

RepeatAlignment<-function(seq1, seq2, seq1.name="seq1", seq2.name="seq2",
    gendist=NULL, dist.type="hamming", dna.model="K80", wmut=1, windel=3.5,
    wslippage=1.75, exclude.pos=NULL, post.include=FALSE, show.steps=FALSE){
    #forked from https://github.com/lucaferretti/RepeatDistance/blob/
    #master/RepeatDistance.R
    #seq1<-as.character(seq1)
    #seq2<-as.character(seq2)
    stopifnot(class(seq1)=="DNAStringSet")
    stopifnot(class(seq2)=="DNAStringSet")
    stopifnot(length(seq1)>1)
    stopifnot(length(seq2)>1)
    orig.seq1.mat <- as.matrix(seq1)
    orig.seq2.mat <- as.matrix(seq2)
    seq1.mat <- as.matrix(seq1)
    seq2.mat <- as.matrix(seq2)
    if(!is.null(exclude.pos)){
        seq1.red <- exclude_XStringSetPos(seq1, exclude.pos)
        seq2.red <- exclude_XStringSetPos(seq2, exclude.pos)
        seq1.mat <- seq1.mat[, -exclude.pos]
        seq2.mat <- seq2.mat[, -exclude.pos]
    }
    #m <- dim(seq1)[1]
    m <- length(seq1)
    #n <- dim(seq2)[1]
    n <- length(seq2)
    #l <- dim(seq1)[2]
    l <- unique(width(seq1))
    #stopifnot(l==dim(seq2)[2])
    stopifnot(l==unique(width(seq2)))
    if(!is.null(gendist)){
        stopifnot(dim(gendist)[1]==m+n, dim(gendist)[2]==m+n)
    } else {
        #gendist=hamming(t(rbind(seq1,seq2)))
        if(!is.null(exclude.pos)){
            if(dist.type=="hamming"){
                gendist <- repeatR::hamming(c(seq1.red, seq2.red))
            }
            if(dist.type=="dna"){
                gendist <- as.matrix(ape::dist.dna(as.DNAbin(
                    c(seq1.red, seq2.red)), model=dna.model))
            }
        } else {
            if(dist.type=="hamming"){
                gendist <- hamming(c(seq1, seq2))
            }
            if(dist.type=="dna"){
                gendist <- as.matrix(ape::dist.dna(as.DNAbin(
                    c(seq1, seq2)), model = dna.model))
            }
        }
    }
    dm <- (gendist[1:m, (m+1):(m+n)] + t(gendist[(m+1):(m+n), 1:m]))/2
    ds <- gendist[1:m, 1:m]
    dt <- gendist[(m+1):(m+n), (m+1):(m+n)]
    d <- matrix(0, nrow=m+1, ncol=n+1)
    mfrom <- matrix(rep(list(), (m+1)*(n+1)), nrow=m+1, ncol=n+1)
    d[2, 1] <- windel
    mfrom[2, 1] <- list(c(3))
    if(show.steps){
        print("gendist")
        print(gendist)
        print("dm:")
        print(dm)
        print("ds:")
        print(ds)
        print("dt:")
        print(dt)
        print("mfrom:")
        print(mfrom)
        print("d:")
        print(d)
        print(mfrom)
    }
    for(i in c(2:m)){
        s_del <- d[i, 1] + windel
        s_ins <- s_del + 1
        s_delslip <- d[i, 1] + wslippage + wmut * min(ds[i, i-1],
            ds[i, i+1-2*(i==m)])
        s_insslip <- s_delslip + 1
        s_mut <- s_del + 1
        d[i+1, 1] <- min(s_mut, s_ins, s_del, s_insslip, s_delslip)
        mfrom[i+1, 1] <- list(which_min(c(s_mut, s_ins, s_del, s_insslip,
            s_delslip)))
        #1: s_mut 2: s_ins 3: s_del 4: s_insslip 5_ s_delslip
        if(show.steps){
            print("i:")
            print(i)
            print("s_del:")
            print(s_del)
            print("s_ins:")
            print(s_ins)
            print("s_delslip:")
            print(s_delslip)
            print("s_insslip:")
            print(s_insslip)
            print("s_mut:")
            print(s_mut)
            print("mfrom:")
            print(mfrom)
            print("d:")
            print(d)
        }
    }
    d[1, 2] <- windel
    mfrom[1, 2] <- list(c(2))
    if(show.steps){
        print("mfrom:")
        print(mfrom)
        print("d:")
        print(d)
    }
    for(j in c(2:n)){
        s_ins <- d[1, j] + windel
        s_del <- s_ins + 1
        s_insslip <- d[1, j] + wslippage + wmut * min(dt[j, j-1],
            dt[j, j+1-2*(j==n)])
        s_delslip <- s_insslip + 1
        s_mut <- s_ins + 1
        d[1, j+1] <- min(s_mut, s_ins, s_del, s_insslip, s_delslip)
        mfrom[1, j+1] <- list(which_min(c(s_mut, s_ins, s_del, s_insslip,
            s_delslip)))# CONVERT TO LIST
        if(show.steps){
            print("j:")
            print(j)
            print("s_del:")
            print(s_del)
            print("s_ins:")
            print(s_ins)
            print("s_delslip:")
            print(s_delslip)
            print("s_insslip:")
            print(s_insslip)
            print("s_mut:")
            print(s_mut)
            print("mfrom:")
            print(mfrom)
            print("d:")
            print(d)
        }
    }
    for (j in c(1:n)){
        for (i in c(1:m)){
            s_del <- d[i, j+1] + windel # deletion
            s_ins <- d[i+1, j] + windel # an insertion
            s_delslip <- d[i, j+1] + wslippage + wmut * ds[i, i-1]
                # slippage on the other
            s_insslip <- d[i+1, j] + wslippage + wmut * dt[j, j-1]
                # slippage
            s_mut <- d[i, j] + dm[i, j] * wmut  # a substitution
            d[i+1, j+1]<-min(s_mut, s_ins, s_del, s_insslip, s_delslip)
            mfrom[i+1, j+1]<-list(which_min(c(s_mut, s_ins, s_del,
                s_insslip, s_delslip)))
            if(show.steps){
                print("i:")
                print(i)
                print("j:")
                print(j)
                print("s_del:")
                print(s_del)
                print("s_ins:")
                print(s_ins)
                print("s_delslip:")
                print(s_delslip)
                print("s_insslip:")
                print(s_insslip)
                print("s_mut:")
                print(s_mut)
                print("mfrom:")
                print(mfrom)
                print("d:")
                print(d)
            }
        }
    }
    result1 <- c()
    result2 <- c()
    order1 <- c()
    order2 <- c()
    mdist <- d[m+1, n+1]
    if(show.steps){
        print(mdist)
        print(d)
    }
    i <- m+1
    j <- n+1
    while(i>1 | j>1){
        if(show.steps){
            print(mfrom[i,j][[1]])
        }
        if(length(mfrom[i, j][[1]])==1){
            orig <- mfrom[i, j][[1]]
        } else {
            orig <- sample(mfrom[i, j][[1]], 1)
        }
        #print(orig)
        if(orig==1){
            if(post.include){
                result1 <- c((orig.seq1.mat[i-1,]), result1)
                result2 <- c((orig.seq2.mat[j-1,]), result2)
            } else {
                result1 <- c((seq1.mat[i-1,]), result1)
                result2 <- c((seq2.mat[j-1,]), result2)
            }
            order1 <- c(i-1, order1)
            order2 <- c(j-1, order2)
            i <- i-1
            j <- j-1
            if(show.steps){
                print("orig==1")
                print(length(result1))
                print(length(result2))
            }
        } else {
            if (orig==2 | orig==4){
                if(post.include){
                    result1 <- c(rep("-", l), result1)
                    result2 <- c((orig.seq2.mat[j-1,]), result2)
                } else {
                    result1 <- c(rep("-",l-length(exclude.pos)), result1)
                    result2 <- c((seq2.mat[j-1,]), result2)
                }
                order1 <- c(0, order1)
                order2 <- c(j-1, order2)
                j <- j-1
                if(show.steps){
                    print("orig==2 | orig==4")
                    print(length(result1))
                    print(length(result2))
                }
            } else {
                if(post.include){
                    result1 <- c((orig.seq1.mat[i-1,]), result1)
                    result2 <- c(rep("-", l), result2)
                } else {
                    result1 <- c((seq1.mat[i-1,]), result1)
                    result2 <- c(rep("-", l-length(exclude.pos)),
                        result2)
                }
                order1 <- c(i-1, order1)
                order2 <- c(0, order2)
                i <- i-1
                if(show.steps){
                    print("orig==other")
                    print(length(result1))
                    print(length(result2))
                }
            }
        }
    }
    names(result1) <- NULL
    names(result2) <- NULL
    names(order1) <- NULL
    names(order2) <- NULL
    if(show.steps){
        print(result1)
        print(result2)
    }
    result <- rbind(result1, result2)
    rownames(result) <- c(seq1.name, seq2.name)
    result <- Biostrings::DNAStringSet(c(paste(result[1,], collapse=""),
        paste(result[2,], collapse="")))
    names(result) <- c(seq1.name, seq2.name)
    result.dist <- NULL
    if(dist.type=="hamming"){
        result.dist <- ape::dist.dna(as.DNAbin(MSA2dist::dnastring2aln(
            result)), model="N")
    }
    if(dist.type=="dna"){
        result.dist <- ape::dist.dna(as.DNAbin(MSA2dist::dnastring2aln(
            result)), model=dna.model)
    }
    orders <- rbind(order1, order2)
    rownames(orders) <- c("order1", "order2")
    results <- list(gendist=gendist, distance=mdist, order=orders,
        sequence=result, sequence.dist=result.dist)
    return(results)
}
