#' @title splitRepByPattern
#' @name splitRepByPattern
#' @description This function splits a repeat containing \code{DNAStringSet},
#' into sequences (repeat units) given a repeat start pattern and repeat length.
#' @param x \code{DNAStringSet}
#' @param p repeat pattern
#' @param l repeat length in nucleotides
#' @param threads number of parallel threads [default: 1]
#' @import foreach
#' @import doMC
#' @importFrom Biostrings DNAString DNAStringSet AAString AAStringSet
#' readDNAStringSet readAAStringSet writeXStringSet width subseq translate
#' @importFrom ape as.DNAbin
#' @importFrom pegas nuc.div
#' @seealso \code{\link[Biostrings]{translate}}
#' @examples
#' ##load example sequence data
#' data("mousePRDM9", package="repeatR")
#' myRepPattern<-"PY"
#' myRepLength<-84
#' mousePRDM9.random<-sample(mousePRDM9, 20)
#' mousePRDM9.random.split<-repeatR::splitRepByPattern(mousePRDM9.random,
#'     myRepPattern, myRepLength)
#' mousePRDM9.random.split$cds
#' mousePRDM9.random.split$aa
#' ##get abundance of repeat units
#' sort(table(as.character(repeatR::collapse_XStringSetList(
#'     mousePRDM9.random.split$aa))))
#' ##get -1, +3, +6 position
#' myStart<-c(37,46,55)
#' myEnd<-c(39,48,57)
#' mousePRDM9.random.split.varPos<-MSA2dist::subString(
#'     repeatR::collapse_XStringSetList(mousePRDM9.random.split$cds),
#'     myStart, myEnd)
#' ##get nucleotide diversity
#' pegas::nuc.div(ape::as.DNAbin(mousePRDM9.random.split.varPos), TRUE)
#' @export splitRepByPattern
#' @author Kristian K Ullrich

splitRepByPattern<-function(x, p, l, threads=1){
    doMC::registerDoMC(threads)
    out <- NULL
    out.cds <- NULL
    out.aa <- NULL
    i <- NULL
    out.cds <- foreach::foreach(i=seq(from=1, to=length(x)),
        .combine=c) %dopar% {
        tmp <- x[i]
        tmp.id <- strsplit(names(tmp)," ")[[1]][1]
        tmp.start <- ((gregexpr(p, as.character(Biostrings::translate(tmp)))
            [[1]])*3)-2
        tmp.cds <- NULL
        tmp.aa <- NULL
        tmp.names <- NULL
        for(s in tmp.start){
            e <- s + l - 1
            if(e > width(tmp)){
                #e <- width(tmp)
                next()
            }
            tmp.cds <- rbind(tmp.cds, as.character(Biostrings::subseq(
                tmp, s, e)))
            tmp.names <- c(tmp.names, paste0(tmp.id, "_", s, "_", e))
        }
        tmp.cds <- Biostrings::DNAStringSet(tmp.cds)
        names(tmp.cds) <- tmp.names
        return(list(tmp.cds))
    }
    names(out.cds) <- unlist(lapply(strsplit(names(x)," "),function(x) x[1]))
    out.aa <- lapply(out.cds, Biostrings::translate)
    out <- list(out.cds, out.aa)
    names(out) <- c("cds", "aa")
    return(out)
}
